from functools import wraps
from typing import Tuple, List
import logging
import os

import numpy as np
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import regularizers, layers, models, optimizers

from .detector import Detector
from ..utils import compute_overlap
from .. import core
from ..utils import get_file, get_datadir_base
from .utils import convert_darknet_weights

"""YOLOv3 Detector

The model architecture here is taken directly from
[this previous work](https://github.com/qqwweee/keras-yolo3). I would
have simply used the original if it were a pip-installable package. The
prediction logic and loss functions are modified to match my stylistic
preferences for documentation, and I take responsibility for any
implementation errors.

The following are a few variable definitions for reference:

- tx, ty, tw, th     : The bounding box outputs from the base model
- txs, tys, twe, the : The bounding box outputs from the training model
- bx, by, bw, bh     : The bounding box outputs from the prediction model
- sx, sy             : The stride of a given box in an output (dependent on the
                       model architecture).
- cx, cy             : The prior offset of a prediction box
- pw, ph             : The prior width and height of a prediction box

In all cases, x and y refer to the center of a given bounding box.

The following formulas govern the relationship between the above variables.
- txs = (bx - cx) / sx
- tys = (by - cy) / sy
- twe = bw / pw
- the = bh / ph
- txs = sigmoid(tx)
- tys = sigmoid(ty)
- twe = exp(tw)
- the = exp(th)

To clarify, txs and tys are factors of the stride to adjust the bounding box.
So txs == 1 corresponds with moving the box center one full stride to the right
from its original position.

Similarly, twe is a factor of the prior width. So twe == 2 corresponds with
doubling the width of the original bounding box.

The main difference between YOLO and Tiny YOLO is that that Tiny YOLO only
uses two scales (two anchor groups with three anchors in each) while YOLO uses
three scales (three anchor groups with three anchors in each)
"""

log = logging.getLogger(__name__)


YOLO_ANCHOR_GROUPS = {
    'tiny': [
        np.array([[81, 82], [135, 169], [344, 319]]),
        np.array([[10, 14], [23, 27], [37, 58]])
    ],
    'full': [
        np.array([[116, 90], [156, 198], [373, 326]]),
        np.array([[30, 61], [62, 5], [59, 119]]),
        np.array([[10, 13], [16, 30], [33, 23]])
    ]
}

YOLO_FILE_CONFIG = {
    'tiny': {
        'weights_fname': 'yolov3-tiny.weights',
        'weights_url': 'https://pjreddie.com/media/files/yolov3-tiny.weights',
        'weights_hash': 'dccea06f59b781ec1234ddf8d1e94b9519a97f4245748a7d4db75d5b7080a42c',   # noqa: E501
        'cfg_fname': 'yolov3-tiny.cfg',
        'cfg_url': 'https://raw.githubusercontent.com/pjreddie/darknet/master/cfg/yolov3-tiny.cfg',  # noqa: E501
        'cfg_hash': '84eb7a675ef87c906019ff5a6e0effe275d175adb75100dcb47f0727917dc2c7',  # noqa: E501
        'converted_fname': 'yolov3_tiny.h5',
        'converted_notop_fname': 'yolov3_tiny_notop.h5',
    },
    'full': {
        'weights_fname': 'yolov3.weights',
        'weights_url': 'https://pjreddie.com/media/files/yolov3.weights',
        'weights_hash': '523e4e69e1d015393a1b0a441cef1d9c7659e3eb2d7e15f793f060a21b32f297',  # noqa: E501
        'cfg_fname': 'yolov3.cfg',
        'cfg_url': 'https://raw.githubusercontent.com/pjreddie/darknet/master/cfg/yolov3.cfg',  # noqa: E501
        'cfg_hash': '22489ea38575dfa36c67a90048e8759576416a79d32dc11e15d2217777b9a953',  # noqa: E501
        'converted_fname': 'yolov3.h5',
        'converted_notop_fname': 'yolov3_notop.h5',
    }
}


@wraps(layers.Conv2D)
def DarknetConv2D(*args, **kwargs):
    """Wrapper to set Darknet parameters for Convolution2D.
    All arguments passed to `layers.Conv2D`"""

    defaults = {
        'kernel_regularizer': regularizers.l2(5e-4),
        'padding': 'valid' if kwargs.get('strides') == (2, 2) else 'same',
    }
    kwargs = {**defaults, **kwargs}
    return layers.Conv2D(*args, **kwargs)


def DarknetConv2D_BN_Leaky(*args, **kwargs):
    """Darknet Convolution2D followed by BatchNormalization and LeakyReLU.
    All arguments passed to DarknetConv2D"""
    defaults = {
        'use_bias': False
    }
    kwargs = {**defaults, **kwargs}

    def make_layer(x):
        x = DarknetConv2D(*args, **kwargs)(x)
        x = layers.BatchNormalization()(x)
        x = layers.LeakyReLU(alpha=0.1)(x)
        return x
    return make_layer


def resblock_body(x, num_filters, num_blocks, name):
    """A series of resblocks starting with a downsampling
    Convolution2D.

    Args:
        x: Input layer
        num_filters: number of filters for DarknetConv2D
        num_blocks: Number of blocks to add
    """
    # Darknet uses left and top padding instead of 'same' mode
    x = layers.ZeroPadding2D(((1, 0), (1, 0)))(x)
    x = DarknetConv2D_BN_Leaky(
        filters=num_filters,
        kernel_size=(3, 3),
        strides=(2, 2),
        name=name+'_c2d_bn_leaky'
    )(x)
    for i in range(num_blocks):
        y = DarknetConv2D_BN_Leaky(
            filters=num_filters // 2,
            kernel_size=(1, 1),
            name=name+'_block{0}_1x1_c2d_bn_leaky'.format(i)
        )(x)
        y = DarknetConv2D_BN_Leaky(
            filters=num_filters,
            kernel_size=(3, 3),
            name=name+'_block{0}_3x3_c2d_bn_leaky'.format(i)
        )(y)
        x = layers.Add(name=name+'_block{0}_add'.format(i))([x, y])
    return x


def make_input_layer(x, num_filters, name):
    xi = DarknetConv2D_BN_Leaky(filters=num_filters, kernel_size=(1, 1), name=name+'_c2d_bn_leaky1')(x)  # noqa: E501
    xi = DarknetConv2D_BN_Leaky(filters=num_filters*2, kernel_size=(3, 3), name=name+'_c2d_bn_leaky2')(xi)  # noqa: E501
    xi = DarknetConv2D_BN_Leaky(filters=num_filters, kernel_size=(1, 1), name=name+'_c2d_bn_leaky3')(xi)  # noqa: E501
    xi = DarknetConv2D_BN_Leaky(filters=num_filters*2, kernel_size=(3, 3), name=name+'_c2d_bn_leaky4')(xi)  # noqa: E501
    xi = DarknetConv2D_BN_Leaky(filters=num_filters, kernel_size=(1, 1), name=name+'_c2d_bn_leaky5')(xi)  # noqa: E501
    x = DarknetConv2D_BN_Leaky(filters=num_filters*2, kernel_size=(3, 3), name=name+'_c2d_bn_leaky_out')(xi)  # noqa: E501
    return x, xi


def get_backbone(inputs, size):
    """Create YOLO_V3 model CNN body in Keras"""
    assert size in ['tiny', 'full'], 'Size must be `tiny` or `full`'
    if size == 'tiny':
        x = DarknetConv2D_BN_Leaky(filters=16, kernel_size=(3, 3))(inputs)
        x = layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same')(x)  # noqa: E501
        x = DarknetConv2D_BN_Leaky(filters=32, kernel_size=(3, 3))(x)
        x = layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same')(x)  # noqa: E501
        x = DarknetConv2D_BN_Leaky(filters=64, kernel_size=(3, 3))(x)
        x = layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same')(x)  # noqa: E501
        x = DarknetConv2D_BN_Leaky(filters=128, kernel_size=(3, 3))(x)
        x = layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same')(x)  # noqa: E501
        x = DarknetConv2D_BN_Leaky(filters=256, kernel_size=(3, 3))(x)

        x1 = layers.MaxPooling2D(pool_size=(2, 2), strides=(2, 2), padding='same')(x)  # noqa: E501
        x1 = DarknetConv2D_BN_Leaky(filters=512, kernel_size=(3, 3))(x1)
        x1 = layers.MaxPooling2D(pool_size=(2, 2), strides=(1, 1), padding='same')(x1)  # noqa: E501
        x1 = DarknetConv2D_BN_Leaky(filters=1024, kernel_size=(3, 3))(x1)
        x1 = DarknetConv2D_BN_Leaky(filters=256, kernel_size=(1, 1))(x1)

        x2 = DarknetConv2D_BN_Leaky(filters=128, kernel_size=(1, 1))(x1)
        x2 = layers.UpSampling2D(2)(x2)
        x2 = layers.Concatenate()([x2, x])
        x2 = DarknetConv2D_BN_Leaky(filters=256, kernel_size=(3, 3))(x2)

        x1 = DarknetConv2D_BN_Leaky(filters=512, kernel_size=(3, 3))(x1)
        outputs = [x1, x2]
    else:
        conv1 = DarknetConv2D_BN_Leaky(32, (3, 3), name='conv_bn_leaky1')(inputs)  # noqa: E501
        rb1 = resblock_body(conv1, 64, 1, name='resblock1')
        rb2 = resblock_body(rb1, 128, 2, name='resblock2')
        rb3 = resblock_body(rb2, 256, 8, name='resblock3')
        rb4 = resblock_body(rb3, 512, 8, name='resblock4')
        rb5 = resblock_body(rb4, 1024, 4, name='resblock5')

        x1, x1i = make_input_layer(rb5, 512, name='x1')

        x2 = DarknetConv2D_BN_Leaky(filters=256, kernel_size=(1, 1))(x1i)
        x2 = layers.UpSampling2D(2, name='upsampling_x2')(x2)

        x2 = layers.Concatenate()([x2, rb4])
        x2, x2i = make_input_layer(x2, 256, name='x2')

        x3 = DarknetConv2D_BN_Leaky(filters=128, kernel_size=(1, 1))(x2i)
        x3 = layers.UpSampling2D(2, name='upsampling_x3')(x3)

        x3 = layers.Concatenate()([x3, rb3])
        x3, x3i = make_input_layer(x3, 128, name='x3')
        outputs = [x1, x2, x3]

    return models.Model(inputs, outputs)


def get_top(backbone, num_anchors_per_output, num_classes):
    filters_per_anchor = (num_classes+5)
    out_filters = num_anchors_per_output*filters_per_anchor
    outputs = []
    for i, x in enumerate(backbone.outputs):
        y = DarknetConv2D(
            filters=out_filters,
            kernel_size=(1, 1),
            name='prediction_group_{0}'.format(i+1)
        )(x)
        concatenate = []
        for n in range(num_anchors_per_output):
            # Below are txs, tys, twe, the, tys, and class probabilites
            startIdx = n*filters_per_anchor
            txsi = startIdx + 0
            tysi = startIdx + 1
            twei = startIdx + 2
            thei = startIdx + 3
            si = startIdx + 4
            csi = startIdx + 5
            csj = csi + num_classes

            def apply_activation(y):
                return layers.Concatenate()([
                    layers.Activation('sigmoid')(y[..., txsi:txsi+1]),
                    layers.Activation('sigmoid')(y[..., tysi:tysi+1]),
                    layers.Activation('exponential')(y[..., twei:twei+1]),
                    layers.Activation('exponential')(y[..., thei:thei+1]),
                    layers.Activation('sigmoid')(y[..., si:si+1]),
                    # As per section 2.2 of the paper, we use independent
                    # logits for each class, instead of softmax.
                    layers.Activation('sigmoid')(y[..., csi:csj])
                ])
            group = layers.Lambda(apply_activation)(y)
            concatenate.append(group)
        y = layers.Concatenate(
            name='anchor_group_{0}'.format(i+1)
        )(concatenate)
        outputs.append(y)
    return models.Model(backbone.inputs, outputs)


def anchors_for_shape(input_shape, anchor_groups):
    """Compute the anchors for a given output shape and anchor
    groups.

    Returns:
        List of anchor coordinates. Each entry in the list is
        of shape (steps_y, steps_x, len(seeds), 6). The entries
        are the base x, y, width, height, horizontal stride, and
        vertical stride, respectively.
    """
    image_h, image_w = input_shape
    anchors = []
    for seeds, stride in zip(anchor_groups, [32, 16, 8]):
        steps_y, steps_x = [int(s / stride) for s in input_shape]
        cx = np.reshape(
            np.arange(start=0, stop=steps_x, dtype='float32'),
            (1, steps_x, 1, 1)
        ) * (image_w / steps_x)
        cy = np.reshape(
            np.arange(start=0, stop=steps_y, dtype='float32'),
            (steps_y, 1, 1, 1)
        ) * (image_h / steps_y)
        xywhs = np.zeros((steps_y, steps_x, len(seeds), 6))
        pw = seeds[:, 0:1]
        ph = seeds[:, 1:2]
        sx = image_w / steps_x
        sy = image_h / steps_y
        xywhs[..., 0:1] = cx
        xywhs[..., 1:2] = cy
        xywhs[..., 2:3] = pw
        xywhs[..., 3:4] = ph
        xywhs[..., 4:5] = sx
        xywhs[..., 5:6] = sy
        anchors.append(xywhs)
    return anchors


def make_loss(num_classes):
    num_features = 4+1+num_classes

    def loss(y_true, y_pred):
        """Loss function for YOLO. y_true and y_pred
        have shape (batchSize, y_steps, x_steps, n_features). This
        loss function is used for each output separately.
        """
        num_anchor_seeds = int(y_pred.shape[-1].value / num_features)
        loss_objectness = 0
        loss_coords = 0
        loss_labels = 0

        for j in range(num_anchor_seeds):
            startIdx = j*num_features
            coordsi = startIdx
            coordsj = coordsi + 4
            labelsi = startIdx + 5
            labelsj = labelsi + num_classes
            objectnessi = startIdx + 4

            # For object state, -1 for ignore, 0 for background, 1
            # for object
            coords_true = y_true[:, :, :, coordsi:coordsj]
            labels_true = y_true[:, :, :, labelsi:labelsj]
            object_state_true = y_true[:, :, :, objectnessi]

            coords_pred = y_pred[:, :, :, coordsi:coordsj]
            labels_pred = y_pred[:, :, :, labelsi:labelsj]
            object_state_pred = y_pred[:, :, :, objectnessi]

            # Calculate objectness loss, which applies to
            # all non-ignore anchors
            indices = tf.where(K.not_equal(object_state_true, -1))
            true = tf.gather_nd(object_state_true, indices)
            pred = tf.gather_nd(object_state_pred, indices)
            loss_objectness += K.sum(K.square(true - pred))

            # Calculate coordinate loss which applies only
            # to positive anchors
            indices = tf.where(K.equal(object_state_true, 1))
            true = tf.gather_nd(coords_true, indices)
            pred = tf.gather_nd(coords_pred, indices)
            loss_coords += K.sum(K.square(true - pred))

            # Calculate classification loss, which applies
            # only to positive anchors. Indices same as above.
            true = tf.gather_nd(labels_true, indices)
            pred = tf.gather_nd(labels_pred, indices)
            loss_labels += K.sum(K.square(true - pred))

        # Calculate total loss
        return loss_objectness + loss_coords + loss_labels

    return loss


def convert_anchors(anchors_xywhs):
    """Convert anchors from xc, yc, w, h to
    x1, y1, x2, y2"""
    anchors_xyxy = np.zeros(
        (len(anchors_xywhs), 4)
    )
    anchors_xyxy[:, 0:2] = anchors_xywhs[:, 0:2] - (anchors_xywhs[:, 2:4] / 2)
    anchors_xyxy[:, 2:4] = anchors_xywhs[:, 0:2] + (anchors_xywhs[:, 2:4] / 2)
    return anchors_xyxy


class YOLOv3(Detector):
    """A detector wrapping
    `YOLOv3 <https://pjreddie.com/media/files/papers/YOLOv3.pdf>`_.
    Most of the implementation is borrowed from
    `keras-yolov3 <https://github.com/qqwweee/keras-yolo3>`_.

    Args:
        annotation_config: The annotation configuration to use for detection
        input_shape: Tuple of (height, width, n_channels)
        size: One of `tiny` or `full`.
        pretrained_backbone: Whether to use a pretrained backbone for the model
        pretrained_top: Whether to use the pretrained full model (only
            supported for where `annotation_config` is
            `AnnotationConfiguration.COCO`)

    Attributes:
        model: The base Keras model containing the weights for feature
            extraction and bounding box regression / classification model.
            You should use this model for loading and saving weights.
        backbone: The backbone for the model (all layers except for final
            classification and regression).
    """
    def __init__(
        self,
        annotation_config: core.AnnotationConfiguration=core.AnnotationConfiguration.COCO,  # noqa: E501
        input_shape: Tuple[int, int]=(None, None, 3),
        size: str='tiny',
        pretrained_backbone: bool=True,
        pretrained_top: bool=False
    ):
        assert (
            not pretrained_top or
            annotation_config == core.AnnotationConfiguration.COCO
        ), (
            'pretrained_top requires annotation configuration '
            'to be AnnotationConfiguration.COCO'
        )
        self.anchor_groups = YOLO_ANCHOR_GROUPS[size]
        self.annotation_config = annotation_config
        self.size = size
        self.backbone = get_backbone(
            inputs=layers.Input(input_shape),
            size=size
        )
        self.model = get_top(
            backbone=self.backbone,
            num_classes=len(annotation_config),
            num_anchors_per_output=len(self.anchor_groups[0])
        )
        if pretrained_backbone and pretrained_top:
            log.warning(
                'pretrained_top makes pretrained_backbone '
                'redundant. Disabling pretrained_backbone.'
            )
            pretrained_backbone = False
        if pretrained_backbone or pretrained_top:
            weights_path = os.path.join(
                get_datadir_base(),
                'weights',
                'yolov3',
                YOLO_FILE_CONFIG[size][
                    'converted_notop_fname' if pretrained_backbone
                    else 'converted_fname'
                ]
            )
            if not os.path.isfile(weights_path):
                log.warning(
                    "Could not find weights file. Attempting to build. "
                    "If error occurs, please execute "
                    "YOLOv3.make_pretrained_weight_files(sizes=['{0}'])".format(  # noqa: E501
                        size
                    ) + " separately first."
                )
                YOLOv3.make_pretrained_weight_files(sizes=[size])
            if pretrained_backbone:
                self.backbone.load_weights(weights_path)
            if pretrained_top:
                self.model.load_weights(weights_path)
        self.compile()

    def invert_targets(
        self,
        y: List[np.ndarray],
        images: List[core.Image],
        threshold: float=0.5
    ) -> core.SceneCollection:
        num_features = 4+1+len(self.annotation_config)
        output_anchors = anchors_for_shape(
            input_shape=images[0].shape[:2],
            anchor_groups=self.anchor_groups
        )
        batch_size = len(y[0])
        processed = []
        for output, anchors in zip(y, output_anchors):
            output = output.reshape((batch_size, -1, num_features))
            anchors = anchors.reshape((1, -1, 6))
            cx, cy, pw, ph, sx, sy = [anchors[:, :, i:i+1] for i in range(6)]
            txs, tys, twe, the, c = [output[:, :, i:i+1] for i in range(5)]
            cc = output[:, :, 5:]
            bx = txs*sx + cx
            by = tys*sy + cy
            bw = pw*twe
            bh = ph*the
            processed.append(np.concatenate([bx, by, bw, bh, c, cc], axis=-1))
        box_groups = [
            np.concatenate([p[idx] for p in processed])
            for idx in range(batch_size)
        ]

        scenes = []
        for boxes, image in zip(box_groups, images):
            boxes = boxes[boxes[:, 4] > threshold]
            annotations = []
            for xi, yi, w, h, s, c in zip(
                boxes[:, 0], boxes[:, 1], boxes[:, 2], boxes[:, 3],
                boxes[:, 4], boxes[:, 5:].argmax(axis=1)
            ):
                annotations.append(
                    core.Annotation(
                        selection=core.Selection(
                            points=[
                                (xi-w/2, yi-h/2),
                                (xi+w/2, yi+h/2),
                            ]
                        ),
                        category=self.annotation_config[c],
                        score=s
                    )
                )
            scenes.append(core.Scene(
                annotation_config=self.annotation_config,
                annotations=annotations,
                image=image
            ))
        return scenes

    def compile(self):
        # Each anchor group has four times as many anchors as the prior one
        n = len(self.anchor_groups)
        total = sum(4**i for i in range(n))
        weights = [4**i / total for i in range(n)]
        self.model.compile(
            loss=make_loss(
                num_classes=len(self.annotation_config)
            ),
            loss_weights=weights,
            optimizer=optimizers.Adam()
        )

    def compute_inputs(self, images: List[core.Image]):
        return np.float32([image.scaled(0, 1) for image in images])

    def compute_targets(self, collection: core.SceneCollection):
        """Preprocess true boxes to training input format

        Args:
            collection: The collection of scenes for which to build output

        Returns:
            y_true: list of array, shape like yolo_outputs, xywh are relative
                values.
        """
        ignore_overlap = 0.5
        anchors = anchors_for_shape(
            input_shape=collection.images[0].shape[:2],
            anchor_groups=self.anchor_groups
        )
        results = [[] for group in anchors]
        anchors_xywh = np.concatenate(
            [a.reshape(-1, 6) for a in anchors],
            axis=0
        )
        anchors_xyxy = convert_anchors(anchors_xywh)
        for scene in collection.scenes:
            annotations = scene.bboxes()
            y_true = np.zeros(
                (len(anchors_xyxy), 4+1+len(self.annotation_config))
            )
            if len(annotations) > 0:
                overlaps = compute_overlap(
                    boxes=anchors_xyxy.astype(np.float64),
                    query_boxes=annotations.astype(np.float64)
                )

                # Set objectness targets
                # The objectness score should be:
                #     1, if the bounding box prior overlaps a ground
                #        truth object by more than any other bounding box prior
                #        and overlaps by more than threshold
                #     0, if the bounding box does not overlap any ground truth
                #        object by more than threshold
                #     ignored (-1), if the bounding box overlaps a ground truth
                #        object by more than threshold, but is not the box of
                #        maximum overlap
                ignored_index = (overlaps >= ignore_overlap).any(axis=1)
                positive_indices = overlaps.argmax(axis=0)
                ignored_index[positive_indices] = False
                background_index = ~ignored_index
                background_index[positive_indices] = False
                y_true[positive_indices, 4] = 1
                y_true[ignored_index, :] = 0
                y_true[ignored_index, 4] = -1
                y_true[background_index, :] = 0
                y_true[background_index, 4] = 0

                # Set class targets
                y_true[positive_indices, 5:] = 0
                y_true[positive_indices, 5+annotations[:, 4]] = 1

                # Set x, y, w, h targets, or more formally,
                # the txs, tys, twe, and the targets (see formulas
                # for additional details).
                bw = annotations[:, 2] - annotations[:, 0]
                bh = annotations[:, 3] - annotations[:, 1]
                bx = annotations[:, 0] + (bw / 2)
                by = annotations[:, 1] + (bh / 2)
                cx = anchors_xywh[positive_indices, 0]
                cy = anchors_xywh[positive_indices, 1]
                pw = anchors_xywh[positive_indices, 2]
                ph = anchors_xywh[positive_indices, 3]
                sx = anchors_xywh[positive_indices, 4]
                sy = anchors_xywh[positive_indices, 5]
                y_true[positive_indices, 0] = (bx - cx) / sx
                y_true[positive_indices, 1] = (by - cy) / sy
                y_true[positive_indices, 2] = bw / pw
                y_true[positive_indices, 3] = bh / ph

            # The loss function should:
            # - Penalize 'positive' (1) boxes for coordinate, objectness,
            #       and class predictions
            # - Penalize 'background' (0) boxes for objectness predictions
            # - Penalize 'ignore' (-1) boxes for nothing
            anchorIdx = 0
            for i, group in enumerate(anchors):
                n_anchors = np.product(group.shape[:3])
                selected = y_true[anchorIdx:anchorIdx+n_anchors].reshape(
                    group.shape[0], group.shape[1], -1
                )
                results[i].append(
                    selected
                )
                anchorIdx += n_anchors
        results = [np.array(r) for r in results]
        return results

    @staticmethod
    def make_pretrained_weight_files(sizes=['tiny', 'full']):
        for size in sizes:
            config = YOLO_FILE_CONFIG[size]
            log.info('Getting pretrained weights')
            converted_path = os.path.join(
                get_datadir_base(),
                'weights',
                'yolov3',
                config['converted_fname']
            )
            converted_path_notop = converted_path.replace(
                config['converted_fname'],
                config['converted_notop_fname']
            )
            if not os.path.isfile(converted_path):
                log.info('Converting darknet weights to keras.')
                log.info('Getting darknet model weights file.')
                weights_fpath = get_file(
                    fname=config['weights_fname'],
                    origin=config['weights_url'],
                    file_hash=config['weights_hash'],
                    cache_subdir=os.path.join('weights', 'yolov3'),
                    hash_algorithm='sha256'
                )
                log.info('Getting model configuration')
                cfg_path = get_file(
                    fname=config['cfg_fname'],
                    origin=config['cfg_url'],
                    file_hash=config['cfg_hash'],
                    cache_subdir=os.path.join('weights', 'yolov3'),
                    hash_algorithm='sha256'
                )
                convert_darknet_weights(
                    config_path=cfg_path,
                    weights_path=weights_fpath,
                    output_path=converted_path,
                    weights_only=True
                )
            else:
                log.info('Found existing converted weights.')
            log.info('Creating notop and complete weight files.')
            backbone = get_backbone(
                inputs=layers.Input((None, None, 3)),
                size=size
            )
            base_model = get_top(
                backbone=backbone,
                num_anchors_per_output=len(YOLO_ANCHOR_GROUPS[size][0]),
                num_classes=len(core.AnnotationConfiguration.COCO)
            )
            base_model.load_weights(converted_path)
            backbone.save_weights(converted_path_notop)


"""
Test dataset functions.
"""

import tempfile
import os

from mira import datasets, core

TEST_VOC_XML_STRING = """
<annotation>
    <folder>VOC2012</folder>
    <filename>test_voc.jpg</filename>
    <source>
        <database>The VOC2007 Database</database>
        <annotation>PASCAL VOC2007</annotation>
        <image>flickr</image>
    </source>
    <size>
        <width>486</width>
        <height>500</height>
        <depth>3</depth>
    </size>
    <segmented>0</segmented>
    <object>
        <name>person</name>
        <pose>Unspecified</pose>
        <truncated>0</truncated>
        <difficult>0</difficult>
        <bndbox>
            <xmin>174</xmin>
            <ymin>101</ymin>
            <xmax>349</xmax>
            <ymax>351</ymax>
        </bndbox>
        <part>
            <name>head</name>
            <bndbox>
                <xmin>169</xmin>
                <ymin>104</ymin>
                <xmax>209</xmax>
                <ymax>146</ymax>
            </bndbox>
        </part>
        <part>
            <name>hand</name>
            <bndbox>
                <xmin>278</xmin>
                <ymin>210</ymin>
                <xmax>297</xmax>
                <ymax>233</ymax>
            </bndbox>
        </part>
        <part>
            <name>foot</name>
            <bndbox>
                <xmin>273</xmin>
                <ymin>333</ymin>
                <xmax>297</xmax>
                <ymax>354</ymax>
            </bndbox>
        </part>
        <part>
            <name>foot</name>
            <bndbox>
                <xmin>319</xmin>
                <ymin>307</ymin>
                <xmax>340</xmax>
                <ymax>326</ymax>
            </bndbox>
        </part>
    </object>
</annotation>
"""


def test_voc():
    """Test VOC loading operations."""
    with tempfile.TemporaryDirectory() as tempdir:
        fpath_xml = os.path.join(tempdir, 'test_voc.xml')
        with open(fpath_xml, 'w') as f:  # pylint: disable=invalid-name
            f.write(TEST_VOC_XML_STRING)
        fpath_jpg = os.path.join(tempdir, 'test_voc.jpg')
        core.Image.new(width=486, height=500, channels=3).save(fpath_jpg)

        original = datasets.load_voc(
            filepaths=[fpath_xml],
            annotation_config=core.AnnotationConfiguration(['person']),
            image_dir=tempdir)
        assert len(original[0].annotations) == 1


def test_via():
    """Test VIA loading and saving operations."""
    collection = datasets.via.load_via('tests/assets/via_project.json')
    with tempfile.TemporaryDirectory() as tempdir:
        savedir = os.path.join(tempdir, 'export')
        datasets.via.save_via(collection, savedir)
        collection_recovered = datasets.via.load_via(
            os.path.join(savedir, 'via.json'))
        assert len(collection_recovered) == 1
        assert len(collection_recovered[0].annotations) == 1
        bbox1 = collection[0].annotations[0].selection.bbox()
        bbox2 = collection_recovered[0].annotations[0].selection.bbox()
        print(bbox1, bbox2)
        assert all(v1 == v2 for v1, v2 in zip(bbox1, bbox2))



import logging
from os import path
import json

import numpy as np

from .. import utils
from ..core import (
    Scene,
    SceneCollection,
    AnnotationConfiguration,
    Annotation,
    Selection
)

log = logging.getLogger(__name__)


def load_coco(
    annotations_file: str,
    image_dir: str,
    annotation_config: AnnotationConfiguration=None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO JSON file.

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    if image_dir is None:
        image_dir = path.split(annotations_file)[0]
    with open(annotations_file, 'r') as f:
        data = json.load(f)
    categories = {}
    for category in data['categories']:
        categories[category['id']] = category['name']
    category_names = [
        c[1] for c in sorted(list(categories.items()), key=lambda x: x[0])
    ]
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(
            category_names
        )
    assert len(categories) == len(annotation_config), \
        'Annotation configuration incompatible with in-file categories'
    assert all([c in annotation_config for c in category_names]), \
        'Some in-file categories not in annotation configuration'
    assert all([c.name in category_names for c in annotation_config]), \
        'Some annotation configuration categories not in file'

    annotations = np.array(
        [
            [ann['image_id'], ann['category_id']] + ann['bbox']
            for ann in data['annotations']]
    )
    annotations = annotations[annotations[:, 0].argsort()]
    images = sorted(data['images'], key=lambda x: x['id'])
    del data
    scenes = [None]*len(images)
    p = utils.Progbar(len(images), task_name='Creating scenes')
    startIdx = 0
    for imageIdx, image in enumerate(images):
        p.update(imageIdx + 1)
        current = annotations[startIdx:][
            annotations[startIdx:, 0] == image['id'], 1:
        ]
        startIdx += len(current)

        scenes[imageIdx] = Scene(
            image=path.join(image_dir, image['file_name']),
            annotation_config=annotation_config,
            annotations=[
                Annotation(
                    category=annotation_config[categories[int(ann[0])]],
                    selection=Selection(
                        points=[
                            [
                                ann[1],
                                ann[2]
                            ],
                            [
                                ann[1] + ann[3],
                                ann[2] + ann[4]
                            ]
                        ]
                    )
                ) for ann in current
            ]
        )
    return SceneCollection(
        scenes=scenes,
        annotation_config=annotation_config
    )


def load_coco_text(
    annotations_file: str,
    image_dir: str,
    annotation_config: AnnotationConfiguration=None,
) -> SceneCollection:
    """Obtain a scene collection from a COCO Text JSON file
    (e.g., that which can be obtained from https://bgshih.github.io/cocotext/)

    Args:
        annotations_file: The annotation file to load
        image_dir: The directory in which to look for images
        annotation_config: The annotation configuration to
            use. If None, it is inferred from the annotations
            category file.

    Returns:
        A scene collection
    """
    with open(annotations_file, 'r') as f:
        data = json.load(f)

    category_names = set([ann['class'] for ann in data['anns'].values()])
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(sorted(list(category_names)))  # noqa: E501
    assert len(category_names) == len(annotation_config), \
        'Annotation configuration incompatible with in-file categories'
    assert all([c in annotation_config for c in category_names]), \
        'Some in-file categories not in annotation configuration'
    assert all([c.name in category_names for c in annotation_config]), \
        'Some annotation configuration categories not in file'

    images = data['imgs']
    p = utils.Progbar(len(images), task_name='Creating scenes')
    scenes = [None]*len(images)

    for sceneIdx, (imageId, imageData) in enumerate(images.items()):
        p.update(sceneIdx + 1)
        anns = [data['anns'][str(annId)] for annId in data['imgToAnns'][imageId]]  # noqa: E501
        scenes[sceneIdx] = Scene(
            annotation_config=annotation_config,
            image=path.join(image_dir, imageData['file_name']),
            annotations=[
                Annotation(
                    category=annotation_config[ann['class']],
                    selection=Selection(
                        [
                            [ann['bbox'][0], ann['bbox'][1]],
                            [ann['bbox'][0] + ann['bbox'][2], ann['bbox'][1] + ann['bbox'][3]]  # noqa: E501
                        ]
                    )
                )
                for ann in anns
            ]
        )
    return SceneCollection(
        scenes=scenes,
        annotation_config=annotation_config
    )


"""
Test core functions.
"""

import tempfile
import os

from mira.core import Image


def test_blank_and_properties():
    """Make sure creating new images works."""
    image = Image.new(width=200, height=100, channels=2, cval=125)
    assert image.height == 100
    assert image.width == 200
    assert image.channels == 2
    assert (image[:, :, :] == 125).all()


def test_file_read():
    """Make sure reading files works."""
    image = Image.new(width=200, height=100, channels=3, cval=125)
    image[40:60, 40:60, 0] = 0
    with tempfile.TemporaryDirectory() as tempdir:
        fpath = os.path.join(tempdir, 'test.jpg')
        image.save(fpath)
        image = Image.read(fpath)

        def consistency_check(img):
            assert img.width == 200
            assert img.height == 100
            assert img.channels == 3
            assert (img[45:55, 45:55, 0] <= 5).all()
            assert (img[45:55, 45:55, 1:] > 100).all()

        consistency_check(image)
        with open(fpath, 'rb') as buffer:
            image = Image.read(buffer)
        consistency_check(image)

        image = Image.read(image.buffer('.jpeg'))
        consistency_check(image)



import sys
import time
import collections
import logging
import os
import hashlib
import io
import shutil
import tarfile
import zipfile
import six
import urllib

import numpy as np
try:
    from keras_retinanet.utils.compute_overlap import compute_overlap  # noqa: F401, E501
except ValueError:
    # Avoids crashing documentation due to numpy
    # warning.
    def compute_overlap():
        pass

log = logging.getLogger(__name__)

# Progbar, _extract_archive, get_file, _hash_file and validate_file
# adapted / copied from Keras to avoid requiring
# TensorFlow to simply use file downloads, etc.

# ProgressFileObject adapted from https://stackoverflow.com/a/3668977


class Progbar(object):
    """Displays a progress bar.

    # Arguments
        target: Total number of steps expected, None if unknown.
        width: Progress bar width on screen.
        verbose: Verbosity mode, 0 (silent), 1 (verbose), 2 (semi-verbose)
        stateful_metrics: Iterable of string names of metrics that
            should *not* be averaged over time. Metrics in this list
            will be displayed as-is. All others will be averaged
            by the progbar before display.
        interval: Minimum visual progress update interval (in seconds).
    """

    def __init__(self, target, width=30, verbose=1, interval=0.05,
                 stateful_metrics=None, task_name=None):
        if task_name is not None:
            print(task_name)
        self.target = target
        self.width = width
        self.verbose = verbose
        self.interval = interval
        if stateful_metrics:
            self.stateful_metrics = set(stateful_metrics)
        else:
            self.stateful_metrics = set()

        self._dynamic_display = ((hasattr(sys.stdout, 'isatty') and
                                  sys.stdout.isatty()) or
                                 'ipykernel' in sys.modules)
        self._total_width = 0
        self._seen_so_far = 0
        self._values = collections.OrderedDict()
        self._start = time.time()
        self._last_update = 0

    def update(self, current, values=None):
        """Updates the progress bar.

        # Arguments
            current: Index of current step.
            values: List of tuples:
                `(name, value_for_last_step)`.
                If `name` is in `stateful_metrics`,
                `value_for_last_step` will be displayed as-is.
                Else, an average of the metric over time will be displayed.
        """
        values = values or []
        for k, v in values:
            if k not in self.stateful_metrics:
                if k not in self._values:
                    self._values[k] = [v * (current - self._seen_so_far),
                                       current - self._seen_so_far]
                else:
                    self._values[k][0] += v * (current - self._seen_so_far)
                    self._values[k][1] += (current - self._seen_so_far)
            else:
                # Stateful metrics output a numeric value.  This representation
                # means "take an average from a single value" but keeps the
                # numeric formatting.
                self._values[k] = [v, 1]
        self._seen_so_far = current

        now = time.time()
        info = ' - %.0fs' % (now - self._start)
        if self.verbose == 1:
            if (now - self._last_update < self.interval and
                    self.target is not None and current < self.target):
                return

            prev_total_width = self._total_width
            if self._dynamic_display:
                sys.stdout.write('\b' * prev_total_width)
                sys.stdout.write('\r')
            else:
                sys.stdout.write('\n')

            if self.target is not None:
                numdigits = int(np.floor(np.log10(self.target))) + 1
                barstr = '%%%dd/%d [' % (numdigits, self.target)
                bar = barstr % current
                prog = float(current) / self.target
                prog_width = int(self.width * prog)
                if prog_width > 0:
                    bar += ('=' * (prog_width - 1))
                    if current < self.target:
                        bar += '>'
                    else:
                        bar += '='
                bar += ('.' * (self.width - prog_width))
                bar += ']'
            else:
                bar = '%7d/Unknown' % current

            self._total_width = len(bar)
            sys.stdout.write(bar)

            if current:
                time_per_unit = (now - self._start) / current
            else:
                time_per_unit = 0
            if self.target is not None and current < self.target:
                eta = time_per_unit * (self.target - current)
                if eta > 3600:
                    eta_format = ('%d:%02d:%02d' %
                                  (eta // 3600, (eta % 3600) // 60, eta % 60))
                elif eta > 60:
                    eta_format = '%d:%02d' % (eta // 60, eta % 60)
                else:
                    eta_format = '%ds' % eta

                info = ' - ETA: %s' % eta_format
            else:
                if time_per_unit >= 1:
                    info += ' %.0fs/step' % time_per_unit
                elif time_per_unit >= 1e-3:
                    info += ' %.0fms/step' % (time_per_unit * 1e3)
                else:
                    info += ' %.0fus/step' % (time_per_unit * 1e6)

            for k in self._values:
                info += ' - %s:' % k
                if isinstance(self._values[k], list):
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if abs(avg) > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                else:
                    info += ' %s' % self._values[k]

            self._total_width += len(info)
            if prev_total_width > self._total_width:
                info += (' ' * (prev_total_width - self._total_width))

            if self.target is not None and current >= self.target:
                info += '\n'

            sys.stdout.write(info)
            sys.stdout.flush()

        elif self.verbose == 2:
            if self.target is None or current >= self.target:
                for k in self._values:
                    info += ' - %s:' % k
                    avg = np.mean(
                        self._values[k][0] / max(1, self._values[k][1]))
                    if avg > 1e-3:
                        info += ' %.4f' % avg
                    else:
                        info += ' %.4e' % avg
                info += '\n'

                sys.stdout.write(info)
                sys.stdout.flush()

        self._last_update = now

    def add(self, n, values=None):
        self.update(self._seen_so_far + n, values)


def get_datadir_base():
    cache_dir_env = os.environ.get('MIRA_CACHE')
    if cache_dir_env is not None:
        cache_dir = os.path.abspath(cache_dir_env)
    else:
        cache_dir = os.path.join(os.path.expanduser('~'), '.MIRA')
    return os.path.expanduser(cache_dir)


class ProgressFileObject(io.FileIO):  # noqa: E302
    def __init__(self, path, *args, **kwargs):
        self._total_size = os.path.getsize(path)
        self.progbar = Progbar(self._total_size, task_name='Reading ' + path)
        io.FileIO.__init__(self, path, *args, **kwargs)

    def read(self, size=-1):
        self.progbar.update(self.tell())
        return io.FileIO.read(self, size)

def _extract_archive(  # noqa: E302
    file_path,
    path='.',
    archive_format='auto',
    extract_check_fn=None
):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == 'auto':
        archive_format = ['tar', 'zip']
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        if archive_type == 'tar':
            # Implement progress bar for tar files,
            # per https://stackoverflow.com/questions/3667865/python-tarfile-progress-output  # noqa: E501
            # open_fn = tarfile.open
            open_fn = lambda fp: tarfile.open(fileobj=ProgressFileObject(fp))  # noqa: E731,E501
            is_match_fn = tarfile.is_tarfile
        if archive_type == 'zip':
            # open_fn = zipfile.ZipFile
            open_fn = lambda fp: zipfile.ZipFile(file=ProgressFileObject(fp))  # noqa: E731,E501
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            if (
                os.path.exists(path) and
                extract_check_fn is not None
            ):
                log.info(
                    'Found target folder. Checking for '
                    'existing complete extraction.'
                )
                if extract_check_fn(path):
                    log.info('Found complete extraction.')
                    return True
                else:
                    log.info('Extraction is not complete. Re-extracting.')
            log.info('Extracting {0} to {1}'.format(file_path, path))  # noqa: E501
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError,
                        KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(
    origin,
    fname=None,
    file_hash=None,
    cache_subdir='datasets',
    hash_algorithm='auto',
    extract=False,
    archive_format='auto',
    extract_check_fn=None
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location. If None,
            the name of the file at origin is used instead.
        origin: Original URL of the file.
        untar: Deprecated in favor of 'extract'.
            boolean, whether the file should be decompressed
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        Path to the downloaded file
    """
    if fname is None:
        fname = os.path.basename(
            urllib.parse.urlparse(origin).path
        )
    datadir_base = get_datadir_base()
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                log.warning(
                    'A local file was found, but it seems to be '
                    'incomplete or outdated because the ' + hash_algorithm +
                    ' file hash does not match the original value of ' +
                    file_hash + ' so we will re-download the data.'
                )
                download = True
    else:
        download = True

    if download:
        class ProgressTracker(object):
            # Maintain progbar for the lifetime of download.
            # This design was chosen for Python 2.7 compatibility.
            progbar = None

        def dl_progress(count, block_size, total_size):
            if ProgressTracker.progbar is None:
                if total_size == -1:
                    total_size = None
                ProgressTracker.progbar = Progbar(
                    total_size,
                    task_name='Downloading data from ' + origin
                )
            else:
                ProgressTracker.progbar.update(count * block_size)

        error_msg = 'URL fetch failure on {} : {} -- {}'
        try:
            try:
                urllib.request.urlretrieve(origin, fpath, dl_progress)
            except urllib.error.HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except urllib.error.URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
        ProgressTracker.progbar = None

    if extract:
        extract_target = os.path.join(
            datadir,
            os.path.split(fpath)[1].split(os.extsep)[0]
        )
        extract_result = _extract_archive(
            file_path=fpath,
            path=extract_target,
            archive_format=archive_format,
            extract_check_fn=extract_check_fn
        )
        if extract_result:
            fpath = extract_target

    return fpath


def _hash_file(fpath, algorithm='sha256', chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    # Example

    ```python
        >>> from keras.data_utils import _hash_file
        >>> _hash_file('/path/to/file.zip')
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        The file hash
    """
    if (algorithm == 'sha256') or (algorithm == 'auto' and len(hash) == 64):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, 'rb') as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b''):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm='auto', chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        Whether the file is valid
    """
    if ((algorithm == 'sha256') or
            (algorithm == 'auto' and len(file_hash) == 64)):
        hasher = 'sha256'
    else:
        hasher = 'md5'
    log.info('Checking hash for {0}.'.format(os.path.split(fpath)[1]))
    if str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash):
        return True
    else:
        return False



"""
Tools for loading and saving VGG Image Annotator datasets.
"""

import os
import json
import uuid
import typing

import pandas as pd

from .. import core


def load_via(project_file: str,
             label_key: str = 'class',
             image_dir: str = None,
             cache: typing.Union[str, bool] = True) -> core.SceneCollection:
    """Load annotations created by the `VGG Image Annotation Tool
    <http://www.robots.ox.ac.uk/~vgg/software/via/>`_.

    Args:
        project_file: The file containing the
            annotations. Only the JSON
            format is supported.
        label_key: The attribute name to use for class labels
        image_dir: The directory in which to look
            for images. If None, defaults the
            folder in which the annotations_file
            is stored.
        cache: The cache parameter to pass to `Scene`. Useful
            if the annotations file references URLs (i.e.,
            specify a directory in which to save images).
    """
    if image_dir is None:
        image_dir = os.path.split(project_file)[0]
    if not project_file.endswith('.json'):
        raise NotImplementedError('Only the VIA JSON format is supported.')
    with open(project_file) as f:  # pylint: disable=invalid-name
        project_data = json.loads(f.read())
    img_metadata = project_data['_via_img_metadata'].values()
    project_attrs = project_data['_via_attributes']
    filenames = [img['filename'] for img in img_metadata]
    regions_df = pd.concat([
        pd.io.json.json_normalize(img['regions']).assign(
            filename=img['filename']
        ) for img in img_metadata
    ], axis=0)

    annotation_config = core.AnnotationConfiguration(
        project_attrs['region'][label_key]['options'].keys()
    )
    assert (regions_df['shape_attributes.name'] == 'rect').all(), \
        'Only axis aligned rectangular selections are supported.'
    return core.SceneCollection(
        scenes=[
            core.Scene(
                image=os.path.join(image_dir, filename),
                annotation_config=annotation_config,
                cache=cache,
                annotations=[
                    core.Annotation(
                        selection=core.Selection(
                            points=[
                                [r['shape_attributes.x'], r['shape_attributes.y']],
                                [
                                    r['shape_attributes.x'] + r['shape_attributes.width'],
                                    r['shape_attributes.y'] + r['shape_attributes.height']
                                ]
                            ]
                        ),
                        category=annotation_config[r[f'region_attributes.{label_key}']]
                    )
                    for _, r in regions_df[regions_df.filename == filename].iterrows()
                ]
            ) for filename in filenames
        ],
        annotation_config=annotation_config
    )

def save_via(
        collection: core.SceneCollection,
        export_dir: str,
        label_key='class',
        project_name='mira_export'
):
    """Save a scene collection in VIA format.

    Args:
        collection: The scene collection to save.
        export_dir: The directory where the via.json file along with all images
            will be saved.
        label_key: The region attribute in which to store the labels
    """
    # Make sure the directory does not exist.
    if os.path.isdir(export_dir):
        raise ValueError(f'{export_dir} already exists.')
    os.makedirs(export_dir, exist_ok=False)
    data = [(f'{uuid.uuid4()}.jpg', scene) for scene in collection]
    for filename, scene in data:
        scene.image.save(os.path.join(export_dir, filename))
    data = [
        (filename, scene, os.path.getsize(os.path.join(export_dir, filename)))
        for filename, scene in data
    ]
    sc_bboxes = [[ann.selection.xywh() for ann in s.annotations] for s in collection]
    img_metadata_keys = [f'{filename}{size}' for filename, _, size in data]
    img_metadata_values = [
        {
            'filename': filename,
            'size': size,
            "file_attributes": {},
            'regions': [
                {
                    'shape_attributes': {
                        'name': 'rect',
                        'x': x, 'y': y, 'width': width, 'height': height
                    },
                    'region_attributes': {
                        label_key: ann.category.name
                    }
                } for (x, y, width, height), ann in zip(bboxes, scene.annotations)
            ]
        } for (filename, scene, size), bboxes in zip(data, sc_bboxes)
    ]
    via_attributes = {
        'region': {
            label_key: {
                'type': 'dropdown',
                'description': '',
                'options': dict(
                    zip(
                        [category.name for category in collection.annotation_config],
                        ["" for category in collection.annotation_config]
                    )
                ),
                'default_options': {}
            }
        },
        'file': {}
    }
    export_object = {
        '_via_img_metadata': dict(zip(img_metadata_keys, img_metadata_values)),
        '_via_attributes': via_attributes,
        "_via_settings": {
            "ui": {
                "annotation_editor_height": 25,
                "annotation_editor_fontsize": 0.8,
                "leftsidebar_width": 18,
                "image_grid": {
                    "img_height": 80,
                    "rshape_fill": "none",
                    "rshape_fill_opacity": 0.3,
                    "rshape_stroke": "yellow",
                    "rshape_stroke_width": 2,
                    "show_region_shape": True,
                    "show_image_policy": "all"
                },
                "image": {
                    "region_label": "__via_region_id__",
                    "region_color": "__via_default_region_color__",
                    "region_label_font": "10px Sans",
                    "on_image_annotation_editor_placement": "NEAR_REGION"
                }
            },
            "core": {
                "buffer_size": 18,
                "filepath": {},
                "default_filepath": ""
            },
            "project": {
                "name": project_name
            }
        }
    }
    with open(os.path.join(export_dir, 'via.json'), 'w') as f:  # pylint: disable=invalid-name
        f.write(json.dumps(export_object))



from .preloaded import (  # noqa: F401
    load_voc2012, load_oxfordiiitpets, load_icdar2015, load_shapes)
from .coco import load_coco  # noqa: F401
from .voc import load_voc  # noqa: F401
from .via import load_via, save_via  # noqa: F401


from .image import Image  # noqa: F401
from .annotation import Annotation, AnnotationConfiguration  # noqa: F401
from .scene import Scene, SceneCollection  # noqa: F401
from .selection import Selection  # noqa: F401



from typing import List, Tuple

import numpy as np

from ..utils import compute_overlap
from ..core import SceneCollection


def precision_recall_curve(
    true_collection: SceneCollection,
    pred_collection: SceneCollection,
    iou_threshold: float=0.5
) -> List[Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Compute the precision-recall curve for each of the
    classes.

    Args:
        true_collection: The true scene collection
        pred_collection: The predicted scene collection
        iou_threshold: The threshold for detection

    Returns:
        A list of 3-tuples, one for each class. The tuple
        consists of the precision, recall, and score
        thresholds, respectively.
    """
    assert true_collection.annotation_config == pred_collection.annotation_config, (  # noqa: E501
        'Annotation configurations must match'
    )
    annotation_config = true_collection.annotation_config
    assert len(true_collection.scenes) == len(pred_collection.scenes), \
        'Must have same scenes in each collection'

    # The ith entry in tfs is a list of length three,
    # which are the change in the number of true positives and
    # false positives, along with the score at which the change
    # occurred for the ith class.
    tfs = [[] for c in range(len(annotation_config))]

    # The ith entry in tfs is the number of true boxes
    # for the ith class.
    pos = [0 for c in range(len(annotation_config))]

    for true, pred in zip(true_collection, pred_collection):
        pred_bboxes = pred.bboxes()
        true_bboxes = true.bboxes()
        pred_scores = pred.scores()

        for c in range(len(annotation_config)):
            pred_bboxes_cur = pred_bboxes[pred_bboxes[:, 4] == c]
            true_bboxes_cur = true_bboxes[true_bboxes[:, 4] == c]
            n = len(pred_bboxes_cur)
            m = len(true_bboxes_cur)
            pos[c] += m

            if n == 0 and m == 0:
                continue

            # (n, m): The IoU between the ith predicted
            #         box and the jth true box
            det = compute_overlap(
                boxes=pred_bboxes_cur.astype('float64'),
                query_boxes=true_bboxes_cur.astype('float64')
            ) > iou_threshold

            # (n): The index of the ith highest confidence
            #      prediction
            pred_bbox_idx = (-pred_scores).argsort()
            tp = 0
            fp = 0
            for i in range(n):
                # For each of the best i bounding boxes,
                # find the number of true ground boxes that
                # were detected.
                tpc = det[pred_bbox_idx[:i+1], :].max(axis=0).sum()
                fpc = (i + 1) - tpc
                sc = pred_scores[pred_bbox_idx[i]]

                tpd = tpc - tp
                fpd = fpc - fp

                tp = tpc
                fp = fpc

                # We keep a running delta for
                # later use.
                tfs[c].append([tpd, fpd, sc])

    prs = [(None, None, None)]*len(annotation_config)

    for tfs_cur, pos_cur, c in zip(
        tfs, pos, range(len(annotation_config))
    ):
        # If we had no detections or there
        # were no true boxes, precision and recall
        # are not defined.
        if len(tfs_cur) == 0 or pos_cur == 0:
            continue
        tfs_cur = np.array(tfs_cur)
        tfs_cur = tfs_cur[(-tfs_cur[:, 2]).argsort()]

        tp = tfs_cur[:, 0].cumsum()
        fp = tfs_cur[:, 1].cumsum()
        prs[c] = (
            tp / (tp + fp),
            tp / pos_cur,
            tfs_cur[:, 2]
        )
    return prs


def mAP(
    true_collection: SceneCollection,
    pred_collection: SceneCollection,
    iou_threshold: float=0.5
) -> float:
    """Compute mAP (mean average precision) for
    a pair of scene collections.

    Args:
        true_collection: The true scene collection
        pred_collection: The predicted scene collection
        iou_threshold: The threshold for detection

    Returns:
        mAP score
    """
    prs = precision_recall_curve(
        true_collection,
        pred_collection,
        iou_threshold
    )
    aps = np.zeros(len(prs)) * np.nan
    for c, (ps, rs, _) in enumerate(prs):
        pi = np.zeros(11)
        for i, r in enumerate(np.linspace(0, 1, 11)):
            # From section 4.2 of VOC paper,the precision at each
            # recall level r is interpolated by taking the maximum
            # precision measured for a method for which
            # the corresponding recall exceeds r
            pc = ps[rs > r]
            if len(pc) > 0:
                pi[i] = pc.max()
        aps[c] = pi.mean()
    return aps.mean()


import warnings

warnings.filterwarnings(
    action="ignore",
    message=r'.*The output shape of `ResNet50\(include_top=False\)` has been changed since Keras 2\.2\.0\..*',  # noqa: E501
    category=UserWarning
)